# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2024 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sqlalchemy.ext.declarative
from sqlalchemy import func, text
from sqlalchemy.types import UserDefinedType
from sqlalchemy.engine import Engine, reflection
import contextlib
import datetime
import calendar
import logging
import sys
import os
try:
    import ConfigParser as configparser
except ImportError:
    import configparser
from syntool_metadata.pattern import (
    OP_OR,
    OP_AND,
    OP_NOT,
    or_ as _pattern_or,
    exec_ast as exec_pattern_ast,
)

logger = logging.getLogger(__name__)
Base = sqlalchemy.ext.declarative.declarative_base()

PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str
else:
    string_types = basestring  # noqa


class UnknownGlobalExtent(Exception):
    def __init__(self, product_id):
        self.product_id = product_id


class MissingProductsTable(Exception):
    pass


# Mappings
# -----------------------------------------------------------------------------
class Geometry(UserDefinedType):
    # Disable statements using this type from being cached at all without a
    # warning
    cache_ok = False

    def get_col_spec(self):
        return "GEOMETRY"

    def bind_expression(self, bindvalue):
        return func.ST_GeomFromText(bindvalue, type_=self)

    def column_expression(self, col):
        return func.AsText(col, type_=self)


class Product(Base):
    """ """
    __tablename__ = 'products'
    product_id = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255),
                                   primary_key=True)
    shortname = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255),
                                  nullable=False)
    type = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255), nullable=False)

    def __init__(self):
        """ """
        self.type = 'ZXY'


class Dataset():
    """ """
    dataset_name = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255),
                                     nullable=False, primary_key=True)
    relative_path = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255),
                                      nullable=False)
    begin_datetime = sqlalchemy.Column(sqlalchemy.types.DATETIME,
                                       nullable=False)
    end_datetime = sqlalchemy.Column(sqlalchemy.types.DATETIME, nullable=False)
    min_zoom_level = sqlalchemy.Column(sqlalchemy.types.INTEGER,
                                       nullable=False)
    max_zoom_level = sqlalchemy.Column(sqlalchemy.types.INTEGER,
                                       nullable=False)
    resolutions = sqlalchemy.Column(sqlalchemy.types.TEXT, nullable=False)
    bbox_text = sqlalchemy.Column(sqlalchemy.types.TEXT, nullable=False)
    bbox_geometry = sqlalchemy.Column(Geometry, nullable=True)
    shape_text = sqlalchemy.Column(sqlalchemy.types.TEXT, nullable=False)
    shape_geometry = sqlalchemy.Column(Geometry, nullable=False)


# Storage implementation
# -----------------------------------------------------------------------------
@sqlalchemy.event.listens_for(Engine, "connect")
def set_mysql_time_zone(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("SET time_zone = '+0:0'")
    cursor.close()


class Storage(object):
    """ """

    def __init__(self, db_url, *args, **kwargs):
        """ """
        super(Storage, self).__init__(*args, **kwargs)
        self.product_tables = {}
        self.__engine = sqlalchemy.create_engine(db_url, echo=False,
                                                 pool_recycle=10)
        self.Base = sqlalchemy.ext.declarative.declarative_base()

    def list_tables(self):
        """"""
        insp = reflection.Inspector.from_engine(self.__engine)
        existing_tables = insp.get_table_names()
        return existing_tables

    def get_product_table(self, product_id):
        fixed_product_id = product_id.replace(' ', '_')
        if fixed_product_id not in self.product_tables:
            table_name = 'product_{}'.format(fixed_product_id)
            product_table = type('Dataset_{}'.format(fixed_product_id),
                                 (self.Base, Dataset),
                                 {'__tablename__': table_name})
            self.product_tables[fixed_product_id] = product_table
        return self.product_tables[fixed_product_id]

    def query(self, *args, **kwargs):
        """"""
        return self.__session.query(*args, **kwargs)

    @contextlib.contextmanager
    def get_session(self):
        """ """
        Session = sqlalchemy.orm.sessionmaker(bind=self.__engine)
        self.__session = Session()
        try:
            yield self
            self.__session.commit()
        except Exception:
            self.__session.rollback()
            raise
        finally:
            self.__session.close()
            logger.info('Session closed')


def get_product_table_name(product_id):
    return 'product_{}'.format(product_id.replace(' ', '_'))


def bbox_polygon_from_text(bbox_text):
    if bbox_text is None:
        return None

    west, south, east, north = bbox_text.split(',')
    bbox_polygon = 'POLYGON(({} {}, {} {}, {} {}, {} {}, {} {}))'
    bbox_polygon = bbox_polygon.format(west, north, east, north, east,
                                       south, west, south, west, north)
    return bbox_polygon


def filter_from_pattern(dataset_cls, pattern_ast):
    return exec_pattern_ast(pattern_ast, {
        '': dataset_cls.dataset_name.like,
        OP_OR: sqlalchemy.or_,
        OP_AND: sqlalchemy.and_,
        OP_NOT: sqlalchemy.not_,
    })


def build_query(q, dataset_cls, bbox_polygon, pattern):
    """"""
    if bbox_polygon is not None:
        geom_intersect = func.Intersects(dataset_cls.shape_geometry,
                                         func.ST_GeomFromText(bbox_polygon))
        q = q.filter(sqlalchemy.or_(dataset_cls.shape_text == 'POINT(0 0)',
                                    geom_intersect))

    if pattern is not None and 0 < len(pattern) and pattern != '%':
        q = q.filter(filter_from_pattern(dataset_cls, pattern))

    return q


def get_image_info(storage, product_id, dataset_id):
    """ """
    dataset_cls = storage.get_product_table(product_id)
    q = storage.query(dataset_cls.bbox_text)
    q = q.filter(dataset_cls.dataset_name == dataset_id)
    q = q.limit(1)
    result = q.first()
    if result is None:
        return None
    if 'POINT(0 0)' == result[0]:
        is_webmercator = (product_id.startswith('900913_') or
                          product_id.startswith('3857_'))
        if is_webmercator is True:
            west = -20037508.34
            north = 20037508.34
            east = 20037508.34
            south = -20037508.34
        elif product_id.startswith('3413_'):
            west = -5000000
            north = 5000000
            east = 5000000
            south = -5000000
        else:
            raise UnknownGlobalExtent(product_id)
    else:
        point0, _, point2, _, _ = result[0][9:-2].split(',')
        west, north = [float(x) for x in point0.split(' ')]
        east, south = [float(x) for x in point2.split(' ')]
    return west, north, east, south


def update_coverage(store, date):
    """ """
    _year = date.year
    if _year not in store:
        store[_year] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    _month = date.month - 1
    _day = date.day - 1
    store[_year][_month] = store[_year][_month] | (1 << _day)


def get_product_coverage(storage, product_id, bbox_polygon, pattern):
    """"""
    """
    SELECT DISTINCT
        DATE(begin_datetime),
        DATE(end_datetime - INTERVAL 1 SECOND)
    FROM `product_${productId.replace(' ', '_')}`
    """

    dataset_cls = storage.get_product_table(product_id)
    dataset_begin = func.date(dataset_cls.begin_datetime)
    dataset_end = func.date(dataset_cls.end_datetime -
                            text('INTERVAL 1 SECOND'))
    q = storage.query(dataset_begin, dataset_end)
    q = build_query(q, dataset_cls, bbox_polygon, pattern)
    q = q.distinct()

    coverage = {}
    inc = datetime.timedelta(days=1)
    for start, stop in q.all():
        while start < stop:
            update_coverage(coverage, start)
            start = start + inc
        update_coverage(coverage, stop)

    return coverage


def get_coverage(storage, bbox_text, products, patterns):
    """
    [product_index, year, january, ..., 0xFFFFFFFF, product_index, year,
     january, ...]
    """
    if products is None or len(products) == 0:
        return []

    existing_tables = storage.list_tables()

    bbox_polygon = bbox_polygon_from_text(bbox_text)

    result = []
    product_index = 0
    for product_id in products:
        pattern = None
        if patterns is not None:
            pattern = patterns[product_index]

        if get_product_table_name(product_id) in existing_tables:
            coverage = get_product_coverage(storage, product_id, bbox_polygon,
                                            pattern)
        else:
            coverage = {}

        result.append(product_index)
        for year in sorted(coverage):
            result.append(year)
            result.extend(coverage[year])
        result.append(0xFFFFFFFF)
        product_index = product_index + 1

    return result


def format_dataset(product_id, product_type, data_url, data_path, dataset,
                   repr_product_id):
    """"""
    resolutions = []

    # Remove suffix for DB entries used to bypass cross-IDL issues
    dataset_name = dataset.dataset_name
    if dataset_name.endswith('_XIDLfix'):
        dataset_name = dataset_name[:-8]

    dataset_uri = '{}/{}/{}/'.format(data_url, repr_product_id, dataset_name)

    if 'POINT(0 0)' == dataset.bbox_text:
        is_webmercator = (product_id.startswith('900913_') or
                          product_id.startswith('3857_'))
        if is_webmercator is True:
            west = -20037508.34
            north = 20037508.34
            east = 20037508.34
            south = -20037508.34
        elif product_id.startswith('3413_'):
            west = -5000000
            north = 5000000
            east = 5000000
            south = -5000000
        else:
            msg = 'Global extent unknown for {}'.format(product_id)
            raise Exception(msg)
    else:
        point0, _, point2, _, _ = dataset.bbox_text[9:-2].split(',')
        west, north = [float(x) for x in point0.split(' ')]
        east, south = [float(x) for x in point2.split(' ')]

    if 'IMAGE' == product_type:
        resolutions = dataset.resolutions
        uri = '{}imageLayer.png'.format(dataset_uri)
    elif product_type in ['VECTOR_FIELD', 'STREAMLINES']:
        resolutions = dataset.resolutions
        uri = '{}vectorFieldLayer.png'.format(dataset_uri)
    elif 'ZXY' == product_type:
        resolutions = [x.split(':')[1] for x in dataset.resolutions.split(',')]
        uri = '{}tiles.zxy/'.format(dataset_uri)
    elif 'TMS' == product_type:
        resolutions = [x.split(':')[1] for x in dataset.resolutions.split(',')]
        uri = '{}tiles/'.format(dataset_uri)
    else:
        uri = dataset_uri

    return {'productId': product_id,
            'datasetId': '{}-{}'.format(product_id, dataset_name),
            'title': dataset_name,
            'uri': uri,
            'start': dataset.begin_datetime.strftime('%Y-%m-%dT%H:%M:%SZ'),
            'end': dataset.end_datetime.strftime('%Y-%m-%dT%H:%M:%SZ'),
            'mapMaxZoom': "{}".format(dataset.max_zoom_level),
            'mapBounds': {'north': '{}'.format(north),
                          'south': '{}'.format(south),
                          'west': '{}'.format(west),
                          'east': '{}'.format(east)},
            'polygon': dataset.shape_text,
            'resolutions': resolutions}


def get_datasets(storage, data_url, data_path, start, stop, bbox_text,
                 requested_products, patterns_by_product, dates_only,
                 results_limit, product_id_aliases):
    """ """
    existing_tables = storage.list_tables()
    products = [p for p in requested_products
                if get_product_table_name(p) in existing_tables]

    bbox_polygon = bbox_polygon_from_text(bbox_text)

    result = []
    bucket_size = results_limit
    for product_id in products:
        if 0 >= bucket_size:
            break

        q0 = storage.query(Product.type)
        q0 = q0.filter(Product.product_id == product_id)
        product_type = q0.first()[0]

        if patterns_by_product is not None:
            pattern = _pattern_or(patterns_by_product[product_id])
        else:
            pattern = None

        dataset_cls = storage.get_product_table(product_id)
        dataset_begin = dataset_cls.begin_datetime
        dataset_end = dataset_cls.end_datetime

        q1 = storage.query(dataset_cls.dataset_name)
        q1 = q1.filter(dataset_begin < stop)
        q1 = q1.filter(dataset_end > start)
        if dates_only:
            q = storage.query(dataset_begin, dataset_end)
        else:
            q = storage.query(dataset_cls.dataset_name,
                              dataset_cls.resolutions,
                              dataset_cls.begin_datetime,
                              dataset_cls.end_datetime,
                              dataset_cls.max_zoom_level,
                              dataset_cls.bbox_text, dataset_cls.shape_text)
        q = q.filter(dataset_cls.dataset_name.in_(q1))
        q = build_query(q, dataset_cls, bbox_polygon, pattern)
        q = q.order_by(dataset_cls.begin_datetime)
        q = q.limit(bucket_size)

        if dates_only:
            product_datasets = [
                [calendar.timegm(x[0].timetuple()),
                 calendar.timegm(x[1].timetuple())]
                for x in q.all()
            ]
        else:
            # Replace product_id used to build the URL of the representation
            # if there is a matching alias in the configuration
            repr_product_id = product_id
            lowercase_product_id = product_id.lower()
            for alias_suffix, actual_id in product_id_aliases.items():
                if lowercase_product_id.endswith(alias_suffix):
                    repr_product_id = actual_id
                    break

            product_datasets = [format_dataset(product_id, product_type,
                                               data_url, data_path, x,
                                               repr_product_id)
                                for x in q.all()]
        bucket_size = bucket_size - len(product_datasets)
        result.extend(product_datasets)
    return result


def get_product_nearest(storage, product_id, bbox_polygon, pattern,
                        current_date, in_past, in_present, in_future):
    """"""
    """
    -- In the past
    SELECT tmp.delta_seconds as delta_seconds
    FROM (
        SELECT
            (
                TIMESTAMPDIFF(SECOND, end_datetime, ${current_date}) +
                TIMESTAMPDIFF(SECOND, begin_datetime, end_datetime) / 2
            ) as delta_seconds
        FROM `product_${productId.replace(' ', '_')}`
        WHERE begin_datetime < ${current_date}
    ) as tmp
    WHERE delta_seconds >= ${in_present ? 0 : 1}
    ORDER BY delta_seconds
    LIMIT 1

    -- In the future
    SELECT tmp.delta_seconds as delta_seconds
    FROM (
        SELECT
            (
                TIMESTAMPDIFF(SECOND, ${current_date}, begin_datetime) +
                TIMESTAMPDIFF(SECOND, begin_datetime, end_datetime) / 2
            ) as delta_seconds
        FROM `product_${productId.replace(' ', '_')}`
        WHERE end_datetime > ${current_date}
    ) as tmp
    WHERE delta_seconds >= ${in_present ? 0 : 1}
    ORDER BY delta_seconds
    LIMIT 1

    -- When not geographically filtering, this modified version gives 2x pref
    SELECT MIN(tmp.delta_seconds) as delta_seconds
    FROM (...) as tmp
    WHERE delta_seconds >= ${in_present ? 0 : 1}
    """

    dataset_cls = storage.get_product_table(product_id)
    unit = text('SECOND')
    duration = func.timestampdiff(unit, dataset_cls.begin_datetime, dataset_cls.end_datetime)
    half_duration = duration / 2

    current_ts = calendar.timegm(current_date.timetuple())
    future_ts = None
    past_ts = None

    if in_past:
        delta_seconds = (
            func.timestampdiff(unit, dataset_cls.end_datetime, current_date) +
            half_duration
        )
        delta_seconds = delta_seconds.label('delta_seconds')
        q = storage.query(delta_seconds)
        q = q.filter(dataset_cls.begin_datetime < current_date)
        q = build_query(q, dataset_cls, bbox_polygon, pattern)
        q = q.subquery()

        delta_seconds = q.c.delta_seconds
        if bbox_polygon is None:
            q = storage.query(func.min(delta_seconds))
        else:
            q = storage.query(delta_seconds)
        q = q.filter(delta_seconds >= (0 if in_present else 1))
        if bbox_polygon is not None:
            q = q.order_by(delta_seconds)
            q = q.limit(1)
        result = q.first()
        if result is not None and result[0] is not None:
            past_ts = current_ts - int(result[0])

    if in_future:
        delta_seconds = (
            func.timestampdiff(unit, current_date, dataset_cls.begin_datetime) +
            half_duration
        )
        delta_seconds = delta_seconds.label('delta_seconds')
        q = storage.query(delta_seconds)
        q = q.filter(dataset_cls.end_datetime > current_date)
        q = build_query(q, dataset_cls, bbox_polygon, pattern)
        q = q.subquery()

        delta_seconds = q.c.delta_seconds
        if bbox_polygon is None:
            q = storage.query(func.min(delta_seconds))
        else:
            q = storage.query(delta_seconds)
        q = q.filter(delta_seconds >= (0 if in_present else 1))
        if bbox_polygon is not None:
            q = q.order_by(delta_seconds)
            q = q.limit(1)
        result = q.first()
        if result is not None and result[0] is not None:
            future_ts = current_ts + int(result[0])

    return past_ts, future_ts


def get_nearest(storage, bbox_text, products, patterns, current_date, in_past,
                in_present, in_future):
    """ """

    result = {}
    if in_past:
        result['past'] = []
    if in_future:
        result['future'] = []

    if products is None or len(products) == 0 or not (in_past or in_future):
        return result

    existing_tables = storage.list_tables()
    bbox_polygon = bbox_polygon_from_text(bbox_text)

    product_index = 0
    for product_id in products:
        pattern = None
        if patterns is not None:
            pattern = patterns[product_index]

        if get_product_table_name(product_id) in existing_tables:
            nearest_past_ts, nearest_future_ts = get_product_nearest(
                storage, product_id, bbox_polygon, pattern,
                current_date, in_past, in_present, in_future
            )
        else:
            nearest_past_ts, nearest_future_ts = None, None

        if in_past:
            result['past'].append(nearest_past_ts)
        if in_future:
            result['future'].append(nearest_future_ts)
        product_index = product_index + 1

    return result


def list_products(storage, **kwargs):
    existing_tables = storage.list_tables()
    if 'products' not in existing_tables:
        raise MissingProductsTable()
    q = storage.query(Product.product_id, Product.shortname)
    results = {k: v for k, v in q.all()}
    return results


def search_datasets(storage, start, stop, bbox_text, requested_products,
                    patterns):
    """ """
    existing_tables = storage.list_tables()
    products = [p for p in requested_products
                if get_product_table_name(p) in existing_tables]

    bbox_polygon = bbox_polygon_from_text(bbox_text)

    results = {}

    product_index = 0
    for product_id in products:
        pattern = None
        if patterns is not None:
            pattern = patterns[product_index]

        dataset_cls = storage.get_product_table(product_id)
        q = storage.query(dataset_cls.dataset_name)
        q = build_query(q, dataset_cls, bbox_polygon, pattern)
        q = q.filter(dataset_cls.begin_datetime < stop)
        q = q.filter(dataset_cls.end_datetime > start)
        q = q.order_by(dataset_cls.begin_datetime)
        results[product_id] = [x[0] for x in q.all()]
        product_index = product_index + 1
    return results


def get_granule_info(data_path, product_id, dataset_name):
    """"""
    features = []
    features_dir = os.path.join(data_path, product_id, dataset_name,
                                'features')
    if os.path.isdir(features_dir):
        for root, dirs, files in os.walk(features_dir):
            ini_features = [os.path.join(root, fn) for fn in files
                            if fn.lower().endswith('.ini')]
            for ini_path in ini_features:
                ini_parser = configparser.ConfigParser()
                ini_parser.read(ini_path)
                feature = dict(ini_parser._sections)
                for k in feature:
                    feature[k].pop('__name__', None)
                features.append(feature)
    return features
