# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2024 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import logging
import syntool_metadata.transect
import syntool_metadata.extract
import syntool_metadata.validation as validation

logger = logging.getLogger(__name__)


def search_datasets(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'minDate', 'maxDate', 'products', 'format',)
    validation.require_params(kwargs, required)

    output_format = kwargs.get('format', 'url')
    if output_format not in ['url', 'path']:
        raise validation.ParameterValidationError(
            'Output format "{}" is not supported, use either "url" or "path"'
            .format(output_format)
        )

    if 'url' == output_format and 'SERVER_HOST' not in environ:
        raise validation.ParameterValidationError(
            'Webserver misconfigured: missing SERVER_HOST environment variable'
        )

    extent = kwargs.get('extent', None)
    if extent is not None and 1 > len(extent):
        extent = None

    products_filter = kwargs.get('products', None)
    if products_filter is not None:
        if 1 > len(products_filter):
            products_filter = None
        else:
            products_filter = products_filter.split(',')

    patterns = kwargs.get('patterns', None)
    if patterns is not None and 0 < len(patterns):
        patterns = patterns.split(',')
    else:
        patterns = None

    min_date = validation.get_date_param(kwargs, 'minDate')
    max_date = validation.get_date_param(kwargs, 'maxDate')

    storage_type = kwargs['webservice_cfg']['storage_type']
    with storage_type.get_session() as storage:
        datasets = storage.search_datasets(min_date,
                                           max_date,
                                           extent,
                                           products_filter,
                                           patterns)

    result = []
    if 'url' == output_format:
        server_host = environ['SERVER_HOST']
        if not server_host.endswith('/'):
            server_host = '{}/'.format(server_host)
        for product_id, product_items in datasets.items():
            product_url = '{}data/ingested/{}'.format(server_host, product_id)
            product_datasets = ['{}/{}'.format(product_url, x)
                                for x in product_items]
            result.extend(product_datasets)
    elif 'path' == output_format:
        for product_id, product_items in datasets.items():
            product_datasets = [os.path.join(product_id, x)
                                for x in product_items]
            result.extend(product_datasets)

    return '\n'.join(result)


def transect(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'productId', 'datasetId', 'x1', 'y1', 'x2',
                'y2', 'zoomLevel',)
    validation.require_params(kwargs, required)

    dataset_id, dataset_type = kwargs['datasetId'].split('+', 1)

    x1 = validation.get_float_param(kwargs, 'x1')
    y1 = validation.get_float_param(kwargs, 'y1')
    x2 = validation.get_float_param(kwargs, 'x2')
    y2 = validation.get_float_param(kwargs, 'y2')
    zoom_level = validation.get_positive_int_param(kwargs, 'zoomLevel')

    ingested_dir = kwargs['webservice_cfg']['ingested_dir']
    result = syntool_metadata.transect.get(ingested_dir,
                                           kwargs['productId'],
                                           dataset_id,
                                           dataset_type,
                                           x1, y1, x2, y2, zoom_level)

    return result


def extract(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'format', 'datasets', 'x1', 'y1', 'x2', 'y2',
                'resolution',)
    supported_formats = ('NUMPY', 'PNG',)
    validation.require_params(kwargs, required)

    output_format = kwargs['format'].upper()
    if output_format not in supported_formats:
        raise validation.ParameterValidationError(
            'unsupported format "{}", expected one of {}'
            .format(output_format, ', '.join(supported_formats))
        )

    if 'SERVER_HOST' not in environ:
        raise validation.ParameterValidationError(
            'Webserver misconfigured: missing SERVER_HOST environment variable'
        )

    x1 = validation.get_float_param(kwargs, 'x1')
    y1 = validation.get_float_param(kwargs, 'y1')
    x2 = validation.get_float_param(kwargs, 'x2')
    y2 = validation.get_float_param(kwargs, 'y2')
    resolution = validation.get_float_param(kwargs, 'resolution')

    storage_type = kwargs['webservice_cfg']['storage_type']
    ingested_dir = kwargs['webservice_cfg']['ingested_dir']
    download_dir = kwargs['webservice_cfg']['download_dir']
    with storage_type.get_session() as storage:
        results = syntool_metadata.extract.get(ingested_dir,
                                               kwargs['datasets'].split(','),
                                               x1, y1, x2, y2, resolution,
                                               output_format,
                                               environ['SERVER_HOST'],
                                               download_dir,
                                               storage)

    return results
