# -*- encoding: utf-8 -*-

"""
@author: <sylvain.gerard@oceandatalab.com>
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2024 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import datetime
import PIL
import PIL.Image
import math
try:
    import simplejson as json
except ImportError:
    import json
import errno
import numpy
import tarfile
import logging
import tempfile
import pyproj.transformer
try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess
import xml.etree.ElementTree
import syntool_metadata.db

logger = logging.getLogger(__name__)


class ProductNotAvailableException(Exception):
    pass


class TilemapNotFound(Exception):
    def __init__(self, tile_dir):
        self.tile_dir = tile_dir


def sign(x):
    """ """
    if x > 0.0:
        return 1.0
    else:
        return -1.0


def clamp(minimum, x, maximum):
    """ """
    return max(minimum, min(x, maximum))


def get_zxy_info(root_dir, product_id, dataset_id, min_resolution):
    """ """
    product_dir = os.path.join(root_dir, product_id)
    dataset_dir = os.path.join(root_dir, product_id, dataset_id)
    tile_dir = os.path.join(dataset_dir, 'tiles.zxy')
    metadata_file = os.path.join(dataset_dir, 'metadata.json')
    json_tile_file = os.path.join(tile_dir, 'tilemap.json')
    xml_tile_file = os.path.join(tile_dir, 'tilemap.xml')

    if not os.path.exists(product_dir):
        raise ProductNotAvailableException(product_id)

    if os.path.exists(metadata_file):
        with open(metadata_file, 'r') as json_file:
            json_data = json.load(json_file)
            point0, _, point2, _, _ = json_data['bbox_str'][9:-2].split(',')
            res_couples = [y.split(':') for y in json_data['resolutions']]
            resolutions = {int(x[0]): float(x[1]) for x in res_couples
                           if x[0] not in ('9998', '9999')}
            epsg = json_data['syntool_id'].split('_', 1)[0]

        if epsg in ['900913', '3857']:
            world_origin = numpy.array([-20037508.34, 20037508.34])
        elif '3413' == epsg:
            world_origin = numpy.array([-5000000, 11384000])
        else:
            raise Exception('Unsupported epsg: {}'.format(epsg))

        west, north = [float(x) for x in point0.split(' ')]
        east, south = [float(x) for x in point2.split(' ')]
        width = east - west
        height = north - south
        origin = numpy.array([west, south])
        for zoom_level, resolution in resolutions.items():
            if resolution <= min_resolution:
                min_resolution = resolution
                break

    elif os.path.exists(json_tile_file):
        with open(json_tile_file, 'r') as json_file:
            json_data = json.load(json_file)
        world_origin = numpy.array(json_data["tiles"]["origin"])

        min_x, min_y, max_x, max_y = json_data["bbox"]
        origin = numpy.array([min_x, min_y])
        width = max_x - min_x
        height = max_y - min_y

        for zoom_level in sorted(json_data["tilesets"], key=lambda x: int(x)):
            resolution = json_data['tilesets'][zoom_level]["units_per_pixel"]
            zoom_level = int(zoom_level)
            if resolution <= min_resolution:
                min_resolution = resolution
                break

    elif os.path.exists(xml_tile_file):
        tree = xml.etree.ElementTree.parse(xml_tile_file)
        root = tree.getroot()
        origin_node = root.findall(".//Origin")[0]
        world_origin_x = float(origin_node.attrib['x'])
        world_origin_y = float(origin_node.attrib['y'])
        world_origin = numpy.array([world_origin_x, world_origin_y])
        bbox_node = root.findall(".//BoundingBox")[0]
        min_x = float(bbox_node.attrib['minx'])
        max_x = float(bbox_node.attrib['maxx'])
        min_y = float(bbox_node.attrib['miny'])
        max_y = float(bbox_node.attrib['maxy'])
        origin = numpy.array([min_x, min_y])
        width = max_x - min_x
        height = max_y - min_y
        tile_set_nodes = root.findall(".//TileSet")
        for tile_set_node in tile_set_nodes:
            resolution = float(tile_set_node.attrib["units-per-pixel"])
            zoom_level = int(tile_set_node.attrib["order"])
            if resolution <= min_resolution:
                min_resolution = resolution
                break
    else:
        raise TilemapNotFound(tile_dir)

    zxy_info = {'product_id': product_id,
                'dataset_id': dataset_id,
                'type': 'ZXY',
                'resolution': resolution,
                'zoom_level': zoom_level,
                "origin": origin,
                'world_origin': world_origin,
                'tile_dir': tile_dir,
                'width': width,
                'height': height}
    return zxy_info, min_resolution


def get_image_info(root_dir, product_id, dataset_id, min_resolution, storage):
    """ """
    west, north, east, south = syntool_metadata.db.get_image_info(storage,
                                                                  product_id,
                                                                  dataset_id)
    image_dir = os.path.join(root_dir, product_id, dataset_id)
    image_path = os.path.join(image_dir, 'imageLayer.png')
    image = PIL.Image.open(image_path, 'r')
    width, height = image.size
    resolution = (east - west) / width
    if resolution < min_resolution:
        min_resolution = resolution

    image_info = {'product_id': product_id,
                  'dataset_id': dataset_id,
                  'type': 'IMAGE',
                  'resolution': resolution,
                  'width': width,
                  'height': height,
                  'west': west,
                  'east': east,
                  'south': south,
                  'north': north,
                  'image_dir': image_dir}
    return image_info, min_resolution


def get_dataset_info(root_dir, min_resolution, product_id, dataset_id,
                     dataset_type, storage):
    """ """
    if 'ZXY' == dataset_type:
        return get_zxy_info(root_dir, product_id, dataset_id, min_resolution)
    elif 'IMAGE' == dataset_type:
        return get_image_info(root_dir, product_id, dataset_id,
                              min_resolution, storage)
    elif 'TMS' == dataset_type:
        raise Exception('TMS not supported anymore')
    raise Exception('"{}" type not supported'.format(dataset_type))


def get_data_from_zxy(dataset, tile_size, min_resolution, x1, y1, x2, y2):
    """ """
    world_origin = dataset["world_origin"]
    resolution = dataset["resolution"]
    zoom_level = dataset["zoom_level"]
    tile_dir = dataset["tile_dir"]
    sx1 = x1 - world_origin[0]
    sy1 = world_origin[1] - y1
    tx1 = sx1 / resolution / tile_size
    ty1 = sy1 / resolution / tile_size
    fracX1, intX1 = math.modf(tx1)
    fracY1, intY1 = math.modf(ty1)
    intX1 = int(intX1)
    intY1 = int(intY1)
    sx2 = x2 - world_origin[0]
    sy2 = world_origin[1] - y2
    tx2 = sx2 / resolution / tile_size
    ty2 = sy2 / resolution / tile_size
    fracX2, intX2 = math.modf(tx2)
    fracY2, intY2 = math.modf(ty2)
    intX2 = int(intX2)
    intY2 = int(intY2)

    dataset_width = (intX2 - intX1 + 1) * tile_size
    dataset_height = (intY1 - intY2 + 1) * tile_size
    dataset_image = None

    i = 0
    j = 0
    for intX in range(intX1, intX2 + 1):
        for intY in range(intY2, intY1 + 1):
            tile_relpath = "%i/%i/%i.png" % (zoom_level, intX, intY)
            tile_path = os.path.join(tile_dir, tile_relpath)
            if os.path.exists(tile_path):
                tile_image = PIL.Image.open(tile_path, "r")
                if dataset_image is None:
                    image_mode = tile_image.mode
                    image_palette = tile_image.palette
                    image_size = (dataset_width, dataset_height)
                    dataset_image = PIL.Image.new(image_mode, image_size)
                dataset_image.paste(tile_image, (i * tile_size, j * tile_size))
            j = j + 1
        i = i + 1
        j = 0

    if dataset_image is None:
        # No tiles for this dataset (totally transparent tiles are not
        # produced).
        return (None, None, None, None, None)

    new_width = int(round(dataset_width * resolution / min_resolution))
    new_height = int(round(dataset_height * resolution / min_resolution))
    new_size = (new_width, new_height)
    dataset_image = dataset_image.resize(new_size, PIL.Image.BILINEAR)
    gx = world_origin[0] + intX1 * tile_size * resolution
    gy = world_origin[1] - intY2 * tile_size * resolution
    image_x = int(round((gx - x1) / min_resolution))
    image_y = int(round((y2 - gy) / min_resolution))
    return (dataset_image, image_x, image_y, image_mode, image_palette)


def get_data_from_image(dataset, min_resolution, x1, y1, x2, y2):
    image_path = os.path.join(dataset["image_dir"], 'imageLayer.png')
    _dataset_image = PIL.Image.open(image_path, "r")
    image_mode = _dataset_image.mode
    image_palette = _dataset_image.palette
    dataset_width = dataset["width"]
    dataset_height = dataset["height"]

    # No need to check that the extraction box intersects image coordinates
    # because the Syntool webapp would not include the dataset in the
    # request if that was the case

    # Compute extraction box coordinates and size in the source image
    # resolution
    x_resolution = abs(dataset['east'] - dataset['west']) / dataset_width
    y_resolution = abs(dataset['north'] - dataset['south']) / dataset_height
    px1 = int(math.floor((x1 - dataset['west']) / x_resolution))
    py1 = int(math.floor((dataset['north'] - y1) / y_resolution))
    px2 = int(math.floor((x2 - dataset['west']) / x_resolution))
    py2 = int(math.floor((dataset['north'] - y2) / y_resolution))
    px1 = min(dataset_width, max(0, px1))
    px2 = min(dataset_width, max(0, px2))
    py1 = min(dataset_height, max(0, py1))
    py2 = min(dataset_height, max(0, py2))

    # Extract relevant area from source image
    img = _dataset_image.crop((px1, py2, px2, py1))

    # Resize extracted area to match expected resolution
    img_size = (abs(px1 - px2), abs(py1 - py2))
    new_width = int(round(img_size[0] * x_resolution / min_resolution))
    new_height = int(round(img_size[1] * y_resolution / min_resolution))
    new_size = (new_width, new_height)
    dataset_image = img.resize(new_size, PIL.Image.BILINEAR)

    # Compute coordinates of the extracted area
    image_x = int(round(max(0, (dataset['west'] - x1) / min_resolution)))
    image_y = int(round(max(0, (y2 - dataset['north']) / min_resolution)))
    return (dataset_image, image_x, image_y, image_mode, image_palette)


def get_dataset_data(dataset, tile_size, min_resolution, x1, y1, x2, y2):
    """ """
    if 'ZXY' == dataset['type']:
        result = get_data_from_zxy(dataset, tile_size, min_resolution, x1, y1,
                                   x2, y2)
        image, image_x, image_y, image_mode, image_palette = result
    elif 'IMAGE' == dataset['type']:
        result = get_data_from_image(dataset, min_resolution, x1, y1, x2, y2)
        image, image_x, image_y, image_mode, image_palette = result
    else:
        raise Exception('"{}" type not supported'.format(dataset['type']))
    return image, image_x, image_y, image_mode, image_palette


def get(root_dir, datasets_list, x1, y1, x2, y2, min_resolution, output_format,
        server_host, download_dir, storage, epsg=3857):
    """ """
    TILE_SIZE = 256
    IMAGE_MAX_SIZE = 8192

    if not server_host.endswith('/'):
        server_host = server_host + '/'

    if x2 < x1:
        x1, x2 = x2, x1

    if y2 < y1:
        y1, y2 = y2, y1

    p1 = numpy.array([x1, y1])
    p2 = numpy.array([x2, y2])

    dx = (p2[0] - p1[0])
    dy = (p2[1] - p1[1])
    ratio = abs(dx / dy)

    # Retrieve dataset info
    # -------------------------------------------------------------------------
    datasets = []
    for dataset in datasets_list:
        product_id, dataset_id, dataset_type = dataset.split("+")
        try:
            dataset_info, min_resolution = get_dataset_info(root_dir,
                                                            min_resolution,
                                                            product_id,
                                                            dataset_id,
                                                            dataset_type,
                                                            storage)
        except ProductNotAvailableException:
            _, e, _ = sys.exc_info()
            logger.debug(e)
            continue
            # Silenced because some product may be stored on a remote server
        datasets.append(dataset_info)

    if not datasets:
        return []

    # Adapt image size
    max_dimension = IMAGE_MAX_SIZE * min_resolution
    if max_dimension < abs(dx) or max_dimension < abs(dy):
        image_width = IMAGE_MAX_SIZE
        image_height = IMAGE_MAX_SIZE
        if ratio < 1.0:
            image_width = IMAGE_MAX_SIZE * ratio
            min_resolution = dy / IMAGE_MAX_SIZE
        else:
            image_height = IMAGE_MAX_SIZE / ratio
            min_resolution = dx / IMAGE_MAX_SIZE
    else:
        image_width = abs(dx) / min_resolution
        image_height = abs(dy) / min_resolution

    image_width = int(round(image_width))
    image_height = int(round(image_height))

    dx = sign(dx) * image_width * min_resolution
    dy = sign(dy) * image_height * min_resolution
    x2 = x1 + dx
    y2 = y1 + dy

    # Prepare output directory
    # -------------------------------------------------------------------------
    if not os.path.exists(download_dir):
        try:
            os.makedirs(download_dir)
        except OSError:
            _, e, _ = sys.exc_info()
            if e.errno != errno.EEXIST:
                raise

    output_dir = tempfile.mkdtemp(dir=download_dir)
    os.chmod(output_dir, 0o755)

    output_dirpath, output_dirname = output_dir, ''
    while output_dirpath and output_dirpath != os.sep and not output_dirname:
        output_dirpath, output_dirname = os.path.split(output_dirpath)

    results = []

    transformer = pyproj.transformer.Transformer.from_crs(epsg, 4326,
                                                          always_xy=True)
    _ = transformer.transform([x1, x2], [y1, y2])
    lon_min, lon_max = _[0]
    lat_min, lat_max = _[1]
    coords_suffix = 'lat_{:.4f}_{:.4f}_lon_{:.4f}_{:.4f}'.format(lat_min,
                                                                 lat_max,
                                                                 lon_min,
                                                                 lon_max)

    # Extract data
    # -------------------------------------------------------------------------
    if 'PNG' == output_format:
        tar_name = 'syntool-archive-{}-{}.tgz'.format(
            datetime.datetime.now().strftime('%Y%m%d_%H%M%S'),
            coords_suffix
        )
        tar_path = os.path.join(output_dir, tar_name)
        tar = tarfile.open(tar_path, "w:gz")
        for dataset in datasets:
            result = get_dataset_data(dataset, TILE_SIZE, min_resolution,
                                      x1, y1, x2, y2)
            image, image_x, image_y, image_mode, image_palette = result
            if image is None:
                continue
            image_size = (image_width, image_height)
            final_image = PIL.Image.new(image_mode, image_size)
            final_image.paste(image, (image_x, image_y))
            if image_mode in ['P', 'L'] and image_palette is not None:
                final_image.putpalette(image_palette.palette)
            final_image_name = '{}-{}.png'.format(dataset['product_id'],
                                                  dataset['dataset_id'])
            final_image_path = os.path.join(output_dir, final_image_name)
            final_image.save(final_image_path)

            new_image_path = os.path.splitext(final_image_path)[0] + ".new.png"
            gdal_opts = ('-q',
                         '-of PNG',
                         '-a_srs "EPSG:{}"'.format(epsg),
                         '-a_ullr {} {} {} {}'.format(x1, y2, x2, y1),
                         '-gcp 0 0 {} {}'.format(x1, y1),
                         '-gcp {} 0 {} {}'.format(image_width, x2, y1),
                         '-gcp 0 {} {} {}'.format(image_height, x1, y2),
                         '-gcp {} {} {} {}'.format(image_width, image_height,
                                                   x2, y2),
                         '"{}"'.format(final_image_path),
                         '"{}"'.format(new_image_path))

            cmd = 'gdal_translate {}'.format(' '.join(gdal_opts))
            subprocess.check_call(cmd, shell=True)
            os.rename(new_image_path, final_image_path)
            os.rename(new_image_path + ".aux.xml",
                      final_image_path + ".aux.xml")

            tar.add(final_image_path, arcname=final_image_name)
            tar.add(final_image_path + ".aux.xml",
                    arcname=final_image_name + ".aux.xml")
            os.unlink(final_image_path)
            os.unlink(final_image_path + ".aux.xml")
        tar.close()
        results.append('{}download/{}/{}'.format(
            server_host,
            output_dirname,
            tar_name
        ))
    elif 'NUMPY' == output_format:
        for dataset in datasets:
            result = get_dataset_data(dataset, TILE_SIZE, min_resolution,
                                      x1, y1, x2, y2)
            image, image_x, image_y, image_mode, image_palette = result
            if image is None:
                continue
            image_size = (image_width, image_height)
            final_image = PIL.Image.new(image_mode, image_size)
            final_image.paste(image, (image_x, image_y))
            array_data = numpy.array(final_image)
            array_name = '{}-{}-{}.npy'.format(dataset['product_id'],
                                               dataset['dataset_id'],
                                               coords_suffix)
            array_path = os.path.join(output_dir, array_name)
            numpy.save(array_path, array_data)

            results.append('{}download/{}/{}'.format(
                server_host,
                output_dirname,
                array_name
            ))
    return results
