# -*- encoding: utf-8 -*-

"""
@author: <ziad.khoury.hanna@oceandatalab.com>

Copyright (C) 2014-2024 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import sys
import re

_PY3 = sys.version_info[0] == 3

if _PY3:
    _STR = str
else:
    _STR = basestring  # noqa


class ParseError(ValueError):
    def __init__(self, issue, pattern, position):
        super(ParseError, self).__init__(
            '{} at index {} in "{}"'.format(issue, position, pattern)
        )

        self.issue = issue
        self.pattern = pattern
        self.position = position


OP_ESCAPE = '\\'
OP_OPEN_PAREN = '('
OP_CLOSE_PAREN = ')'
OP_AND = '&'
OP_OR = '|'
OP_NOT = '!'

_ordered_operators = [OP_OR, OP_AND, OP_NOT]
_LOW_OP = _ordered_operators[0]
_precedence = {op: p for p, op in enumerate(_ordered_operators)}


def _make_operand(op, operands):
    if len(operands) == 1:
        if op != OP_NOT:
            # or(a) <=> a
            # and(aa) <=> aa
            return operands[0]
        if not isinstance(operands[0], _STR) and op == operands[0][0]:
            # not(not(x)) <=> x
            return operands[0][1][0]
        return (op, operands)

    if op != OP_NOT and all(isinstance(o, _STR) or op == o[0] for o in operands):
        # or(a, or(b, c)) <=> or(a, b, c)
        # and(a, and(b, c)) <=> and(a, b, c)
        operands = [
            operand
            for o in operands
            for operand in ((o,) if isinstance(o, _STR) else o[1])
        ]
    return (op, operands)


def or_(operands):
    return _make_operand(OP_OR, operands)


def and_(operands):
    return _make_operand(OP_AND, operands)


def not_(operand):
    return _make_operand(OP_NOT, [operand])


def parse(pattern):
    if pattern and not pattern.startswith('('):
        # legacy pattern
        return '%{}%'.format(pattern)

    stack = []
    operands = []
    op = _LOW_OP
    operand = None
    escaping = False
    buffer = []

    def push_char():
        nonlocal operand, buffer, c, i
        if c == ' ' and not buffer:
            return
        if operand is not None:
            raise ParseError('Unexpected character "{}"'.format(c), pattern, i)
        if c in '\%_':  # noqa
            c = '\\' + c
        if not escaping:
            if c == '*':
                c = '%'
            elif c == '?':
                c = '_'
        buffer.append(c)

    def buffer2operand():
        nonlocal buffer, operand
        operand = ''.join(buffer).rstrip()
        del buffer[:]

    def push_operand():
        nonlocal operand
        if operand is None:
            buffer2operand()
        operands.append(operand)
        operand = None

    def open_parens():
        nonlocal operands, c, i
        if buffer or operand is not None:
            raise ParseError('Unexpected character "{}"'.format(c), pattern, i)
        stack.append((op, operands))
        operands = []

    def close_parens():
        nonlocal operand, operands, op
        operand = _make_operand(op, operands)
        if not stack:
            raise ParseError('Mismatched closing parenthesis', pattern, i)
        op, operands = stack.pop()

    for i, c in enumerate(pattern):
        if escaping:
            push_char()
            escaping = False
        elif c == OP_ESCAPE:
            escaping = True
        elif c == OP_OPEN_PAREN:
            open_parens()
            op = _LOW_OP
        elif c == OP_CLOSE_PAREN:
            push_operand()
            while op != _LOW_OP:
                close_parens()
                push_operand()
            close_parens()
        elif c == OP_OR or c == OP_AND:
            if operand is None:
                buffer2operand()
            # op has higher precedence
            while op != c and _precedence[op] >= _precedence[c]:
                push_operand()
                close_parens()
            # op has lower precedence
            if _precedence[op] < _precedence[c]:
                o, operand = operand, None
                open_parens()
                operand = o
                op = c
            push_operand()
        elif c == OP_NOT:
            open_parens()
            op = c
        else:
            push_char()
    push_operand()
    while stack:
        close_parens()
        push_operand()

    return _make_operand(op, operands)


def exec_ast(ast, operators):
    if isinstance(ast, _STR):
        return operators[''](ast)
    op, operands = ast
    args = (exec_ast(operand, operators) for operand in operands)
    return operators[op](*args)


_wildcard_re = re.compile('((?:\\\\{2})*(?:\\\\[_%])?)[_%]*%[_%]*')
_simplify_operators = {
    '': lambda a: _wildcard_re.sub('\\1%', a),
    OP_OR: lambda *a: '%' if any(x == '%' for x in a) else or_(list(x for x in a if x != '')),
    OP_AND: lambda *a: '' if any(x == '' for x in a) else and_(list(x for x in a if x != '%')),
    OP_NOT: lambda a: '' if a == '%' else ('%' if a == '' else not_(a)),
}


def simplify_ast(ast):
    return exec_ast(ast, _simplify_operators)


if __name__ == '__main__':
    from contextlib import contextmanager
    import traceback

    @contextmanager
    def test(title):
        try:
            print(title)
            yield
            print('✅')
        except:  # noqa
            print('❌ ', end='')
            print(traceback.format_exc())

    with test('Pattern - parse - empty'):
        assert parse('(') == ''
        assert parse('(!') == ('!', [''])
        assert parse('(&') == ('&', ['', ''])
        assert parse('(|') == ('|', ['', ''])
        assert parse('(a&') == ('&', ['a', ''])
        assert parse('(a|') == ('|', ['a', ''])
        assert parse('(&a') == ('&', ['', 'a'])
        assert parse('(|a') == ('|', ['', 'a'])

    with test('Pattern - parse - ignore whitespace around operators'):
        assert parse('( ') == parse('(')
        assert parse('( ! ') == parse('(!')
        assert parse('( & ') == parse('(&')
        assert parse('( | ') == parse('(|')
        assert parse('( a & ') == parse('(a&')
        assert parse('( a | ') == parse('(a|')
        assert parse('( & a ') == parse('(&a')
        assert parse('( | a ') == parse('(|a')
        assert (
            parse('(  ( a & b ) & c | ( ( ( d | e ) & ( f | ! ( g | ! ! h) & ! i & j ')
            == parse('(a&b)&c|(((d|e)&(f|!(g|!!h)&!i&j')
        )

    with test('Pattern - parse - keep whitespace inside atoms'):
        assert parse('(a b') == parse('(a b')

    with test('Pattern - parse - single'):
        assert parse('(a') == 'a'

    with test('Pattern - parse - multi'):
        assert parse('(a | c') == ('|', ['a', 'c'])

    with test('Pattern - parse - unwrap | with single operand'):
        assert parse('(a & b') == ('&', ['a', 'b'])

    with test('Pattern - parse - flatten &, |'):
        assert parse('(c & (d & e)') == ('&', ['c', 'd', 'e'])
        assert parse('(c | (d | e)') == ('|', ['c', 'd', 'e'])

    with test('Pattern - parse - complex'):
        assert parse('(g | !h) & i') == ('&', [('|', ['g', ('!', ['h'])]), 'i'])
        assert (
            parse('(a & b) & c | (((d | e) & (f | !(g | !!h) & !i & j')
            == ('|', [
                ('&', ['a', 'b', 'c']),
                ('&', [
                    ('|', ['d', 'e']),
                    ('|', [
                        'f',
                        ('&', [
                            ('!', [('|', ['g', 'h'])]),
                            ('!', ['i']),
                            'j',
                        ]),
                    ]),
                ]),
            ])
        )

    with test('Pattern - parse - fail'):
        try:
            parse('())')
            assert False
        except ParseError as error:
            assert error.issue == 'Mismatched closing parenthesis'
            assert error.pattern == '())'
            assert error.position == 2

    with test('Pattern - exec'):
        operators = {
            '': _STR,
            OP_OR: lambda *a: '({})'.format(' or '.join(a)),
            OP_AND: lambda *a: '({})'.format(' and '.join(a)),
            OP_NOT: lambda a: 'not {}'.format(a),
        }

        assert (
            exec_ast(
                parse('(a_% & b\\%\\_) & c* | (((d? | e) & (f | !(g | !!h) & !i & j'),
                operators
            )
            == '((a\\_\\% and b\\%\\_ and c%) or ((d_ or e) and (f or (not (g or h) and not i and j))))'
        )
        assert (
            exec_ast(
                parse('(a?* & b\\*\\?) & c* | (((d? | e) & (f | !(g | !!h) & !i & j'),
                operators
            )
            == '((a_% and b*? and c%) or ((d_ or e) and (f or (not (g or h) and not i and j))))'
        )

    with test('Pattern - simplify'):
        assert simplify_ast(parse('a\\_')) == '%a\\_%'
        assert simplify_ast(parse('(**??**?*?*?')) == '%'
        assert simplify_ast(parse('(**?*')) == '%'
        assert simplify_ast(parse('(???')) == '___'
        assert simplify_ast(parse('(?a?*?b')) == '_a%b'
        assert simplify_ast(parse('(?*a?*b?*c?*')) == '%a%b%c%'
        assert simplify_ast(parse('(**?a??')) == '%a__'
        assert simplify_ast(parse('(??b*?*')) == '__b%'
        assert simplify_ast(parse('(??b*?*a?*')) == '__b%a%'
        assert simplify_ast(parse('(!*a*&!*b*')) == ('&', [('!', ['%a%']), ('!', ['%b%'])])
        assert simplify_ast(parse('(**?a?? & ??b*?* & **?* | ')) == ('&', ['%a__', '__b%'])
