# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2024 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import webob
import routr
import logging
import syntool_metadata.webservice as webservice
from syntool_metadata.validation import ParameterValidationError

PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str
else:
    string_types = basestring  # noqa

logger = logging.getLogger(__name__)
webservice_cfg = webservice.load_cfg()
routes = webservice.init_routes(webservice_cfg['url_prefix'],
                                webservice_cfg['full_service'])


def application(environ, start_response):
    """WSGI application"""
    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    request = webob.Request(environ)
    try:
        trace = routes(request)
        view = trace.target
        args, kwargs = trace.args, trace.kwargs
        for param_name, param_value in request.params.items():
            kwargs[param_name] = param_value

        kwargs['webservice_cfg'] = webservice_cfg

        response = view(environ, *args, **kwargs)
    except ParameterValidationError as error:
        response = {'ok': False, 'details': str(error)}
    except routr.exc.NoMatchFound:
        e = sys.exc_info()[1]
        logger.debug('404', exc_info=e)
        response = e.response
    except webob.exc.HTTPException:
        e = sys.exc_info()[1]
        logger.debug('HTTP 40x/50x', exc_info=e)
        response = e

    is_response = isinstance(response, webob.Response)
    if is_response is False:
        is_dict = isinstance(response, dict)
        is_list = isinstance(response, list)
        is_tuple = isinstance(response, tuple)
        is_basestring = isinstance(response, string_types)
        if (is_dict or is_list or is_tuple):
            response = webob.Response(json=response)
        elif is_basestring:
            response = webob.Response(response)
        elif response is None:
            response = webob.Response()
    return response(environ, start_response)
